import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Cron, CronExpression } from '@nestjs/schedule';
import { Model } from 'mongoose';
import { md5 } from '../shared/utils';
import { IMailchimpSubscription } from './interface/mailchimp-subscription';
import { INewsletterSubscription } from './interface/newsletter-subscription.interface';
import { NewsletterSubscription } from './newsletter-subscription.schema';
import mailchimp = require('@mailchimp/mailchimp_marketing');

@Injectable()
export class NewsletterService {
  private readonly logger = new Logger(NewsletterService.name);
  private LIST_ID = process.env.MC_LIST_ID;
  constructor(
    @InjectModel(NewsletterSubscription.name) private newsletterSubscriptionModel: Model<INewsletterSubscription>
  ) {
    // Configure mailchimp client
    mailchimp.setConfig({
      apiKey: process.env.MC_API_KEY,
      server: process.env.MC_SERVER,
    });
  }

  @Cron(CronExpression.EVERY_DAY_AT_3AM)
  public async updateNewsletterSubscription(): Promise<void> {
    this.logger.debug('updateNewsletterSubscription');
    const { total_items } = await mailchimp.lists.getListMembersInfo(this.LIST_ID);
    const { members } = await mailchimp.lists.getListMembersInfo(this.LIST_ID, {
      fields: ['members.email_address,members.id,members.status'],
      count: total_items,
    });
    const memberToRemove = members.filter((user: IMailchimpSubscription) => user.status !== 'subscribed');

    memberToRemove.forEach(async (member: IMailchimpSubscription) => {
      const userSubscription = await this.findOne(member.email_address);
      if (userSubscription) {
        this.logger.log(`Remove subscription ${member.id}`);
        userSubscription.deleteOne();
      }
    });
  }

  public async newsletterSubscribe(email: string): Promise<NewsletterSubscription> {
    this.logger.debug(`newsletterSubscribe: ${email}`);
    email = email.toLocaleLowerCase();
    try {
      // Add or update list member (to be able to subscribe again a member who had already subscribed then unsubscribed)
      // (the second parameter is the MD5 hash of the lowercase email, we actually don't need to maintain a mapping in newsletterSubscription : cf. https://mailchimp.com/developer/marketing/docs/methods-parameters/#path-parameters )
      const member = await mailchimp.lists.setListMember(this.LIST_ID, md5(email), {
        email_address: email,
        status_if_new: 'subscribed', // cf. https://mailchimp.com/developer/marketing/api/list-members/add-or-update-list-member/
        status: 'subscribed',
      });

      // We may not be aware that the user has unsubscribed from the newsletter, so it is ok if it already exists in newsletterSubscription
      let newsletterSubscription = await this.findOne(email);
      if (!newsletterSubscription) {
        newsletterSubscription = await this.newsletterSubscriptionModel.create({
          email: email,
          mailchimpId: member.id,
        });
      }
      return newsletterSubscription;
    } catch (e) {
      if (e.status === 400 && e.response?.text?.includes('fake')) {
        throw new HttpException('Fake or invalid email', HttpStatus.I_AM_A_TEAPOT);
      } else {
        this.logger.error(`newsletterSubscribe ${email}: ${JSON.stringify(e)}`);
        throw new HttpException('Subscribe error', HttpStatus.INTERNAL_SERVER_ERROR);
      }
    }
  }

  public async newsletterUnsubscribe(email: string): Promise<NewsletterSubscription> {
    this.logger.debug(`newsletterUnsubscribe: ${email}`);
    email = email.toLocaleLowerCase();
    const emailMd5Hashed = md5(email);

    let newsletterSubscription = await this.findOne(email);
    if (newsletterSubscription) {
      newsletterSubscription = newsletterSubscription.deleteOne();
    }

    try {
      const response = await mailchimp.lists.getListMember(this.LIST_ID, emailMd5Hashed);
      if (response.status === 'unsubscribed') {
        throw new HttpException('Email not found', HttpStatus.NOT_FOUND);
      }
      await mailchimp.lists.setListMember(this.LIST_ID, emailMd5Hashed, {
        status: 'unsubscribed',
      });
    } catch (e) {
      if (e.status === 404) {
        throw new HttpException('Email not found', HttpStatus.NOT_FOUND);
      } else {
        this.logger.error(`newsletterUnsubscribe ${email}: ${JSON.stringify(e)}`);
        throw new HttpException('Unsubscribe error', HttpStatus.INTERNAL_SERVER_ERROR);
      }
    }

    return newsletterSubscription;
  }

  public async findOne(mail: string): Promise<INewsletterSubscription | undefined> {
    this.logger.debug('findOne');
    return this.newsletterSubscriptionModel.findOne({ email: mail });
  }

  public async findAll(): Promise<NewsletterSubscription[]> {
    this.logger.debug('findAll');
    return this.newsletterSubscriptionModel.find();
  }
}
