import { UsersService } from './../users/services/users.service';
import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Model } from 'mongoose';
import { StructuresService } from '../structures/services/structures.service';
import { CreatePersonalOfferDto } from './dto/create-personal-offer.dto';
import { PersonalOfferDto } from './dto/personal-offer.dto';
import { PersonalOffer, PersonalOfferDocument } from './schemas/personal-offer.schema';

@Injectable()
export class PersonalOffersService {
  private readonly logger = new Logger(PersonalOffersService.name);

  constructor(
    private structuresService: StructuresService,
    private usersService: UsersService,
    @InjectModel(PersonalOffer.name) private personalOfferModel: Model<PersonalOfferDocument>
  ) {}

  public async findOne(id: string): Promise<PersonalOffer> {
    this.logger.debug(`findOne`);
    const result: PersonalOfferDocument = await this.personalOfferModel.findById(id).exec();
    if (!result) {
      throw new HttpException('Personal offer does not exist', HttpStatus.NOT_FOUND);
    }
    return result;
  }

  public async create(createDto: CreatePersonalOfferDto): Promise<PersonalOfferDocument> {
    this.logger.debug(`create`);
    return this.personalOfferModel.create(createDto.personalOffer);
  }

  public async update(id: string, updatePersonalOfferDto: PersonalOfferDto): Promise<PersonalOfferDocument> {
    this.logger.debug(`updating personal offer | ${id}`);
    const result: PersonalOfferDocument = await this.personalOfferModel
      .findByIdAndUpdate(id, updatePersonalOfferDto)
      .exec();
    if (!result) {
      throw new HttpException('Invalid personal offer id for update', HttpStatus.BAD_REQUEST);
    }
    return result;
  }

  /**
   * Delete a personal offer and remove it from both user and structure offers
   */
  public async delete(id: string): Promise<PersonalOfferDocument> {
    this.logger.debug(`delete`);
    const structure = await this.structuresService.findByPersonalOfferId(id);
    this.structuresService.removePersonalOffer(structure, id);

    const user = await this.usersService.findByPersonalOfferId(id);
    this.usersService.removePersonalOffer(user, id);
    const deletedOffer: PersonalOfferDocument = await this.personalOfferModel.findByIdAndDelete(id).exec();
    if (!deletedOffer) {
      throw new HttpException('Invalid personal offer id for deletion', HttpStatus.BAD_REQUEST);
    }
    return deletedOffer;
  }
}
