import { HttpService } from '@nestjs/axios';
import { Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { AxiosResponse } from 'axios';
import * as https from 'https';
import * as _ from 'lodash';
import { Model } from 'mongoose';
import { Observable } from 'rxjs';
import { CategoriesService } from '../../categories/services/categories.service';
import { UsersService } from '../../users/services/users.service';
import { Address } from '../schemas/address.schema';
import { ApticCatalog } from '../schemas/aptic-catalog.schema';
import { ApticStructure } from '../schemas/aptic-structure.schema';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { StructureTypeService } from './../structure-type/structure-type.service';
import { StructuresSearchService } from './structures-search.service';
import { StructuresService } from './structures.service';

@Injectable()
export class ApticStructuresService {
  private readonly logger = new Logger(ApticStructuresService.name);
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    private readonly categoriesService: CategoriesService,
    private structureService: StructuresService,
    private structuresSearchService: StructuresSearchService,
    private structureTypeService: StructureTypeService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  /**
   * Get all aptic structures in the given postal codes.
   * @param postalCodeData
   */
  public formatApticStructures(postalCodeData: any[]): void {
    // Get all postal code in one array
    const postalCodeArray = _.flatten(postalCodeData.map((data) => data.codesPostaux));
    // Call APTIC Api's
    postalCodeArray.map((postalCode) => {
      this.logger.debug(`formatApticStructures | postal code: ${postalCode}`);
      return this.getApticStructures(postalCode).subscribe(
        (res) => {
          res.data.presencePoints.forEach((structure) => {
            this.logger.debug(`formatApticStructures | postal code: ${postalCode} | received : '${structure.name}'`);
            // Call aptic api for offer
            this.getApticStructureOffer(structure.catalogs[0]).subscribe(
              (serviceData) => {
                structure.services = serviceData.data.services;
                // Create structure
                this.createApticStructures(structure);
              },
              (err) => {
                this.logger.log(err);
              }
            );
          });
        },
        (err) => {
          this.logger.log(
            `formatApticStructures | getApticStructures error on postal code: ${postalCode}. Code: ${err}`
          );
        }
      );
    });
  }

  /**
   * Create a structure for app database given an aptic structure
   * @param structure ApticStructure
   */
  private async createApticStructures(structure: ApticStructure): Promise<void> {
    this.structureAlreadyExist(structure).then(async (exist) => {
      if (!exist) {
        this.logger.log(`createApticStructures | Create structure : ${structure.name}`);
        const createdStructure = new this.structureModel();
        createdStructure.categories = {};
        // Known fields
        createdStructure.structureName = structure.name;
        createdStructure.contactPhone = structure.phone;
        // Unkown fields (but mandatory)
        createdStructure.contactMail = null;
        createdStructure.categories.labelsQualifications = ['passNumerique'];
        createdStructure.structureType = await this.structureTypeService.findByValue('autre');
        createdStructure.pmrAccess = false;
        createdStructure.remoteAccompaniment = false;
        createdStructure.categories.accessModality = ['accesLibre'];
        createdStructure.accountVerified = true;
        createdStructure.freeWorkShop = 'Non';
        createdStructure.nbComputers = 0;
        createdStructure.nbPrinters = 0;
        createdStructure.nbScanners = 0;
        // Address
        createdStructure.coord = [structure.address.gpsLng, structure.address.gpsLat];
        createdStructure.address = this.formatAddress(structure);
        // Set structure offer
        createdStructure.categories.selfServiceMaterial = await this.setModules(structure, 'selfServiceMaterial');
        createdStructure.categories.solidarityMaterial = await this.setModules(structure, 'solidarityMaterial');
        createdStructure.categories.onlineProcedures = await this.setModules(structure, 'onlineProcedures');
        createdStructure.categories.baseSkills = await this.setModules(structure, 'baseSkills');
        createdStructure.categories.advancedSkills = await this.setModules(structure, 'advancedSkills');
        createdStructure.categories.accessModality = await this.setModules(structure, 'accessModality');
        createdStructure.categories.labelsQualifications = await this.setModules(structure, 'labelsQualifications');
        createdStructure.categories.age = await this.setModules(structure, 'age');
        createdStructure.categories.languageAndIlliteracy = await this.setModules(structure, 'languageAndIlliteracy');
        createdStructure.categories.handicaps = await this.setModules(structure, 'handicaps');
        createdStructure.categories.publicOthers = ['toutPublic'];
        createdStructure.save();
        this.structuresSearchService.indexStructure(createdStructure);
        // Send admin weird structure mail
        this.userService.sendAdminApticNewStructureMail(createdStructure);
        this.verifyDuplication(createdStructure);
      }
    });
  }

  /**
   * Given an aptic structure, this method return the corresponding services id by category.
   * @param structure ApticStructure
   * @param moduleCategory string
   * @returns Promise<string[]>
   */
  private async setModules(structure: ApticStructure, moduleCategory: string): Promise<string[]> {
    this.logger.debug(`setModules ${moduleCategory} for ${structure.name}`);
    const referentialIds = await this.categoriesService.findByApticId(moduleCategory).then((referential) => {
      return _.flatten(
        // Loop on all modules of a category
        referential.modules.map((repository) => {
          // If there is no aptic link, skip
          if (repository.apticIds !== undefined) {
            // Return the category id corresponding to aptic code and remove duplicate
            return [
              ...new Set(
                repository.apticIds.map((id) => {
                  if (structure.services.find((service) => service.code == id)) {
                    return repository.id;
                  }
                })
              ),
            ];
          } else {
            return [];
          }
        })
      );
    });
    return referentialIds;
  }

  /**
   * Verifiy if an aptic structure already exist in database.
   *  - If it's true return true and update fields if needed
   *  - If it's false return false
   * @param structure ApticStructure
   * @returns boolean
   */
  private async structureAlreadyExist(structure: ApticStructure): Promise<boolean> {
    let existingStructure = await this.structureModel
      .findOne({
        structureName: { $regex: structure.name, $options: 'i' },
      })
      .exec();
    // Check without regex for case like 'TINEBRA*DANIEL/DANIEL/'
    if (!existingStructure) {
      existingStructure = await this.structureModel.findOne({ structureName: structure.name }).exec();
    }

    if (existingStructure) {
      // Add aptic label if it's not the case
      if (!existingStructure.categories.labelsQualifications.includes('passNumerique')) {
        existingStructure.categories.labelsQualifications.push('passNumerique');
      }
      // Update service offer
      existingStructure.categories.selfServiceMaterial = _.unionWith(
        existingStructure.categories.selfServiceMaterial,
        await this.setModules(structure, 'selfServiceMaterial'),
        _.isEqual
      );
      existingStructure.categories.solidarityMaterial = _.unionWith(
        existingStructure.categories.solidarityMaterial,
        await this.setModules(structure, 'solidarityMaterial'),
        _.isEqual
      );
      existingStructure.categories.onlineProcedures = _.unionWith(
        existingStructure.categories.onlineProcedures,
        await this.setModules(structure, 'onlineProcedures'),
        _.isEqual
      );
      existingStructure.categories.baseSkills = _.unionWith(
        existingStructure.categories.baseSkills,
        await this.setModules(structure, 'baseSkills'),
        _.isEqual
      );
      existingStructure.categories.advancedSkills = _.unionWith(
        existingStructure.categories.advancedSkills,
        await this.setModules(structure, 'advancedSkills'),
        _.isEqual
      );
      existingStructure.categories.labelsQualifications = _.unionWith(
        existingStructure.categories.labelsQualifications,
        await this.setModules(structure, 'labelsQualifications'),
        _.isEqual
      );
      existingStructure.categories.age = _.unionWith(
        existingStructure.categories.age,
        await this.setModules(structure, 'age'),
        _.isEqual
      );
      existingStructure.categories.languageAndIlliteracy = _.unionWith(
        existingStructure.categories.languageAndIlliteracy,
        await this.setModules(structure, 'languageAndIlliteracy'),
        _.isEqual
      );
      existingStructure.categories.handicaps = _.unionWith(
        existingStructure.categories.handicaps,
        await this.setModules(structure, 'handicaps'),
        _.isEqual
      );
      existingStructure.save();
      return true;
    }
    return false;
  }

  /**
   * Get Metropole new aptic structure evey week. For testing, please change the expression
   */
  public getMetropoleMunicipality(): void {
    const req =
      'https://download.data.grandlyon.com/ws/grandlyon/adr_voie_lieu.adrcomgl/all.json?maxfeatures=-1&start=1';
    this.logger.log(`getMetropoleMunicipality | Request : ${req}`, '');
    this.httpService.get(encodeURI(req)).subscribe(
      (data) => {
        const inseeArray = data.data.values.map((municipality) => {
          return this.getPostalCodeWithINSEE(municipality.insee).toPromise();
        });
        Promise.all(inseeArray).then((inseData) => {
          const postalCodeArray = inseData.map((cpData: { data; config }) => {
            return cpData.data;
          });
          this.formatApticStructures(postalCodeArray);
        });
      },
      (err) => this.logger.error(err)
    );
  }

  public getPostalCodeWithINSEE(inseeCode: string): Observable<AxiosResponse<unknown>> {
    const req = `https://geo.api.gouv.fr/communes/${inseeCode}?fields=codesPostaux&format=json`;
    this.logger.debug(`getMetropoleMunicipality | Request : ${req}`);
    return this.httpService.get(encodeURI(req));
  }

  public getApticStructures(postalCodeData: string): Observable<AxiosResponse<{ presencePoints: ApticStructure[] }>> {
    const req = `https://aptisearch-api.aptic.fr/v1/postal-code/${postalCodeData}`;
    this.logger.debug(`getApticStructures | Request : ${req}`);
    return this.httpService.get(req, {
      headers: {
        api_key: process.env.APTIC_TOKEN,
      },
    });
  }

  public getApticStructureOffer(catalogId: string): Observable<AxiosResponse<{ services: ApticCatalog[] }>> {
    const agent = new https.Agent({
      rejectUnauthorized: false,
    });
    const req = `https://aptisearch-api.aptic.fr/v1/catalog/${catalogId}/services`;
    this.logger.log(`getApticStructureOffer | Request : ${req}`);
    return this.httpService.get(req, {
      httpsAgent: agent,
      headers: {
        api_key: process.env.APTIC_TOKEN,
      },
    });
  }

  private async verifyDuplication(createdStructure: StructureDocument): Promise<void> {
    const sameAddrStructure = await this.structureModel
      .findOne({
        _id: { $ne: createdStructure._id },
        address: createdStructure.address,
      })
      .exec();
    if (sameAddrStructure) {
      this.userService.sendAdminApticStructureMail(createdStructure.structureName, sameAddrStructure);
    }
  }

  /**
   * Format aptic structure address
   */
  private formatAddress(structure: ApticStructure): Address {
    const address = new Address();
    const regexWithSpace = /\d+\s/g; // NOSONAR
    const regex = /\d+/g; // NOSONAR
    if (structure.address.main.match(regex)) {
      address.numero = structure.address.main.match(regex)[0];
      address.street = structure.address.main.replace(regexWithSpace, '');
    } else {
      address.street = structure.address.main;
    }

    address.commune = this.structureService.getFormattedCity(structure.address.city, structure.address.zip);
    address.postcode = structure.address.zip;
    return address;
  }
}
