import { HttpService } from '@nestjs/axios';
import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Cron, CronExpression } from '@nestjs/schedule';
import { AxiosResponse } from 'axios';
import * as ejs from 'ejs';
import * as _ from 'lodash';
import { DateTime } from 'luxon';
import { DocumentDefinition, FilterQuery, Model, Types } from 'mongoose';
import { lastValueFrom, map, tap } from 'rxjs';
import { PendingStructureDto } from '../../admin/dto/pending-structure.dto';
import { UnclaimedStructureDto } from '../../admin/dto/unclaimed-structure-dto';
import { Categories } from '../../categories/schemas/categories.schema';
import { Module } from '../../categories/schemas/module.class';
import { CategoriesService } from '../../categories/services/categories.service';
import { MailerService } from '../../mailer/mailer.service';
import { PersonalOfferDocument } from '../../personal-offers/schemas/personal-offer.schema';
import { IUser } from '../../users/interfaces/user.interface';
import { User } from '../../users/schemas/user.schema';
import { UsersService } from '../../users/services/users.service';
import { PhotonResponse } from '../common/photonResponse.type';
import { depRegex } from '../common/regex';
import { StructureDto } from '../dto/structure.dto';
import { UpdateStructureDto } from '../dto/update-structure.dto';
import { EquipmentsServicesEnum } from '../enum/equipmentsServices';
import { CNFSStructure } from '../interfaces/cnfs-structure.interface';
import { Address } from '../schemas/address.schema';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { StructuresSearchService } from './structures-search.service';

@Injectable()
export class StructuresService {
  private readonly logger = new Logger(StructuresService.name);
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    private readonly mailerService: MailerService,
    private structuresSearchService: StructuresSearchService,
    private categoriesService: CategoriesService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  async initiateStructureIndex(): Promise<StructureDocument[]> {
    this.logger.log('Reset structures indexes');
    await this.structuresSearchService.dropIndex();
    await this.structuresSearchService.createStructureIndex();
    return this.populateES();
  }

  async searchForStructures(
    text: string,
    filters?: Array<any>,
    fields?: string[],
    limit?: number
  ): Promise<StructureDocument[]> {
    this.logger.debug(
      `searchForStructures : ${text} | filters: ${JSON.stringify(filters)} | fields : ${JSON.stringify(fields)} | `
    );
    const results = await this.structuresSearchService.search(text, fields);
    const ids = results.map((result) => result.structureId);
    let structures;
    const andFilters = filters ? filters[0] : null;
    const orFilters = filters ? filters[1] : null;

    if (!ids.length) {
      return [];
    }

    // we match ids from Elasticsearch with ids from mongoDB (and filters) and sort the result according to ElasticSearch order.
    if (andFilters?.length > 0 && orFilters?.length == 0) {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $and: [...this.parseFilter(andFilters), { deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .populate('structureType')
        .limit(limit)
        .exec();
    } else if (andFilters?.length > 0 && orFilters?.length > 0) {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $or: [...this.parseFilter(orFilters)],
          $and: [...this.parseFilter(andFilters), { deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .populate('structureType')
        .limit(limit)
        .exec();
    } else if (andFilters?.length == 0 && orFilters?.length > 0) {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $or: [...this.parseFilter(orFilters)],
          $and: [{ deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .populate('structureType')
        .limit(limit)
        .exec();
    } else {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $and: [{ deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .populate('structureType')
        .limit(limit)
        .exec();
    }

    return structures.sort((a, b) => ids.indexOf(a.id) - ids.indexOf(b.id));
  }

  public async create(email: string, structure: StructureDto): Promise<Structure> {
    const user = await this.userService.findOne(email);
    if (!user) {
      throw new HttpException('Invalid profile', HttpStatus.NOT_FOUND);
    }
    // Handle equipments
    // Value is forced to true because there is no verification needed now. Cannot be removed because of old data prod
    structure.accountVerified = true;
    const createdStructure = new this.structureModel(structure);
    createdStructure._id = new Types.ObjectId();
    createdStructure.structureName = createdStructure.structureName.trim();
    createdStructure.categories.selfServiceMaterial = this.getSelfServiceMaterial(createdStructure);
    await createdStructure.save();
    await this.getStructurePosition(createdStructure).then(async (structureWithPosition: StructureDocument) => {
      return this.structuresSearchService.indexStructure(
        await this.structureModel
          .findByIdAndUpdate(new Types.ObjectId(createdStructure._id), structureWithPosition, { new: true })
          .populate('personalOffers')
          .populate('structureType')
          .exec()
      );
    });
    user.structuresLink.push(createdStructure._id);
    user.save();

    // Send admin notification mail
    this.sendAdminStructureNotification(
      createdStructure,
      this.mailerService.config.templates.adminStructureCreate.ejs,
      this.mailerService.config.templates.adminStructureCreate.json
    );

    // Async treatment that can be done in background, DO NOT await because it takes around 3sec to finish
    this.bindOneCNFSStructure(createdStructure.id);

    return createdStructure;
  }

  public async search(searchString: string, filters?: Array<any>): Promise<Structure[]> {
    if (searchString && filters) {
      return this.structureModel
        .find({
          $and: [
            ...this.parseFilter(filters),
            { $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true },
          ],
        })
        .populate('personalOffers')
        .populate('structureType')
        .exec();
    } else if (filters) {
      return this.structureModel
        .find({ $and: [{ $and: this.parseFilter(filters), deletedAt: { $exists: false }, accountVerified: true }] })
        .populate('personalOffers')
        .populate('structureType')
        .exec();
    } else {
      return this.structureModel
        .find({
          $and: [{ $or: [{ $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true }] }],
        })
        .populate('personalOffers')
        .populate('structureType')
        .exec();
    }
  }

  /**
   * Parse equipments. Ex : if structure doesn't have computers but nbComputer changes, then the key is added to equipmentsAndServices
   * @param structure
   * @returns Structure
   */
  public getSelfServiceMaterial(structure: StructureDocument): string[] {
    Logger.debug(`getSelfServiceMaterial of ${structure.structureName}`, StructuresService.name);
    if (!structure.categories.selfServiceMaterial) {
      structure.categories.selfServiceMaterial = [];
    }
    if (structure.nbComputers > 0) {
      structure.categories.selfServiceMaterial.push(EquipmentsServicesEnum.computer);
    }
    if (structure.nbPrinters > 0) {
      structure.categories.selfServiceMaterial.push(EquipmentsServicesEnum.printer);
    }
    if (structure.nbScanners > 0) {
      structure.categories.selfServiceMaterial.push(EquipmentsServicesEnum.scanner);
    }
    // // Remove duplicate if exists
    return [...new Set(structure.categories.selfServiceMaterial)];
  }

  /**
   * Parse filter value from string to boolean and map filter category to categories key
   * @param filters
   */
  private parseFilter(filters: Array<{ [key: string]: unknown }>): Array<{ [key: string]: unknown }> {
    return filters.map((filter) => {
      const [key, value] = Object.entries(filter)[0];

      if (typeof value === 'string' && value.toLowerCase() === 'true') {
        return { [key]: true };
      } else {
        return { [`categories.${key}`]: value };
      }
    });
  }

  /**
   * @param deleted boolean to get deleted structures
   */
  public async findAll(deleted = false): Promise<StructureDocument[]> {
    this.logger.debug('findAll');
    return this.structureModel
      .find({ deletedAt: { $exists: deleted }, accountVerified: true })
      .populate('structureType')
      .exec();
  }

  public async findAllUnclaimed(): Promise<UnclaimedStructureDto[]> {
    const structures = await this.structureModel
      .find({ deletedAt: { $exists: false } })
      .populate('personalOffers')
      .populate('structureType')
      .exec();
    const unclaimedStructures = [];
    await Promise.all(
      structures.map(async (structure: StructureDocument) => {
        if (!(await this.userService.isStructureClaimed(structure.id))) {
          unclaimedStructures.push({
            structureId: structure.id,
            structureName: structure.structureName,
            createdAt: structure.createdAt,
            updatedAt: structure.updatedAt,
          });
        }
      })
    );
    return unclaimedStructures;
  }

  public async populateES(): Promise<StructureDocument[]> {
    const structures = await this.structureModel
      .find({ deletedAt: { $exists: false } })
      .populate('personalOffers')
      .populate('structureType')
      .exec();
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        this.structuresSearchService.indexStructure(structure);
      })
    );
    return structures;
  }

  /**
   * Find all structure and format them in order to be directly crawled
   */
  public async findAllFormated(categories: Categories[]): Promise<StructureDocument[]> {
    this.logger.debug('Find all formatted structures, only returning structures who consented to data sharing');
    return (
      await this.structureModel
        .find({ deletedAt: { $exists: false }, accountVerified: true, $where: 'this.dataShareConsentDate != null' })
        .select('-_id -accountVerified -otherDescription -dataShareConsentDate')
        .exec()
    ).map((structure) => {
      const repositoryKeys = categories.map((category) => category.id);
      repositoryKeys.forEach((el) => {
        // Add referentiel
        if (structure.categories[el] !== undefined) {
          structure.categories[el] = this.mapModules(
            structure.categories[el],
            categories.find((category) => category.id === el).modules
          );
        }
      });
      return structure;
    });
  }
  /**
   * Map modules labels to ids
   * @param structureModule
   * @param baseModule
   * @returns
   */
  public mapModules(structureModule: string[], baseModule: Module[]): string[] {
    if (structureModule.length === 0) {
      return [];
    }
    return structureModule.map((id) => {
      const formatedText = baseModule.find((module) => {
        return module.id === id;
      })?.name;
      return formatedText ? formatedText : id;
    });
  }

  public async update(idStructure: string, updatedFields: UpdateStructureDto): Promise<Structure> {
    this.logger.debug(`Updating structure ${idStructure}`);
    const oldStructure = await this.findOne(idStructure);

    // Update position only if address has changed (to not erase coord which may have been set manually in database)
    if (updatedFields.address && !Address.equals(updatedFields.address, oldStructure.address)) {
      oldStructure.address = updatedFields.address;
      await this.getStructurePosition(oldStructure);
    }

    // Store updated categories because it will be erased by Object.assign which is a shallow copy
    const oldCategories = { ...oldStructure.categories };
    const updatedCategories = { ...updatedFields.categories };
    const deepClone: StructureDocument = Object.assign(oldStructure, updatedFields);
    // Update structure categories in order to not override it
    if (updatedCategories) {
      deepClone.categories = Object.assign(deepClone.categories, oldCategories);
      Object.keys(updatedCategories).forEach((key) => {
        deepClone.categories[key] = updatedCategories[key];
      });
    }

    if (updatedFields.structureName) {
      deepClone.structureName = deepClone.structureName.trim();
    }
    const parsedStructure = deepClone;
    parsedStructure.categories.selfServiceMaterial = this.getSelfServiceMaterial(deepClone);

    const result = await this.structureModel.findByIdAndUpdate(new Types.ObjectId(idStructure), parsedStructure).exec();
    if (!result) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const responseStructure = await this.findOne(idStructure);
    await this.structuresSearchService.update(responseStructure, idStructure);
    this.sendAdminStructureNotification(
      result,
      this.mailerService.config.templates.structureModificationNotification.ejs,
      this.mailerService.config.templates.structureModificationNotification.json
    );
    this.userService.removeOutdatedStructureFromArray(idStructure);

    return responseStructure;
  }

  public async findOne(idParam: string): Promise<StructureDocument> {
    this.logger.debug('findOne');
    return this.structureModel
      .findById(new Types.ObjectId(idParam))
      .populate('personalOffers')
      .populate('structureType')
      .exec();
  }
  /**
   * Get structures positions and add marker corresponding to those positons on the map
   */
  private async getStructurePosition(structure: Structure): Promise<Structure> {
    this.logger.debug('getStructurePosition');
    try {
      let address: {
        geometry: {
          coordinates: [number, number];
        };
        properties: {
          [key: string]: string;
        };
      };
      // check if photon-bal is more precise than photon ban
      const photonBalCoord = await this.getCoord(
        structure.address.numero || '',
        structure.address.street,
        structure.address.commune,
        'photon-bal'
      );
      if (photonBalCoord.data.features.length > 0) {
        address = photonBalCoord.data.features.find((feature) => feature.properties.hasOwnProperty('housenumber'));
        if (!address) {
          address = photonBalCoord.data.features[0];
        }
      } else {
        // else pick structure from photonban if it exists or log error
        const photonCoord = await this.getCoord(
          structure.address.numero || '',
          structure.address.street,
          structure.address.commune,
          'photon'
        );
        address = photonCoord.data.features.find((feature) => feature.properties.hasOwnProperty('housenumber'));
      }

      if (!address || !address.geometry) {
        this.logger.error(
          `No coord found for: ${structure.address.numero || ''} ${structure.address.street} ${
            structure.address.commune
          }`,
          StructuresService.name
        );
        structure.coord = [];
      } else {
        structure.coord = address.geometry.coordinates;
        structure.address.postcode = address.properties.postcode;
      }

      // set territory for CTM filter
      await this.setCtmTerritory(structure);
    } catch (err) {
      this.logger.error(err);
    }

    return structure;
  }

  /**
   * Get structure territory
   */
  private async setCtmTerritory(structure: Structure): Promise<Structure> {
    structure.categories['ctm'] = [];
    if (structure.address?.postcode) {
      let inseeCode: string = null;
      const req = `https://apicarto.ign.fr/api/codes-postaux/communes/${structure.address.postcode}`;
      this.logger.debug(`Request : ${req}`);

      const res = await lastValueFrom(this.httpService.get(encodeURI(req)));
      if (res.data?.length) {
        if (res.data.length == 1) {
          this.logger.debug(`Insee code ${res.data[0].codeCommune} for ${structure.address.commune}`);
          inseeCode = res.data[0].codeCommune;
        } else {
          // if several commune for this postcode, find which commune we are looking for the insee code
          const communes = res.data.filter((commune) => commune.nomCommune == structure.address.commune);
          if (communes.length) {
            this.logger.debug(`Insee code ${communes[0].codeCommune} for ${structure.address.commune}`);
            inseeCode = res.data[0].codeCommune;
          } else {
            this.logger.warn(
              `Commune not found for ${structure.address.commune} in ${JSON.stringify(res.data)} (${req})`
            );
          }
        }
      }

      if (inseeCode) {
        structure.address.inseeCode = inseeCode;
        const ctm: Module = await this.categoriesService.findCtmByInseeCode(inseeCode);
        if (ctm) {
          this.logger.debug(`${structure.structureName} : CTM ${ctm.id}`);
          structure.categories['ctm'] = [ctm.id];
        }
      }
    }
    return structure;
  }

  public async isClaimed(structureId: string, user: User): Promise<boolean> {
    const isStructureClaimed = await this.userService.isStructureClaimed(structureId);
    const isUserAlreadyClaimed = await this.userService.isUserAlreadyClaimedStructure(structureId, user.email);
    if (isStructureClaimed || isUserAlreadyClaimed) {
      return true;
    }
    return false;
  }

  /**
   * Get city formatted without "Arrondissement"
   * @param {searchQuery} data - Query address
   */
  public getFormattedCity(city: string, postcode: string) {
    let formattedCity: string = city;
    if (city?.startsWith('Lyon')) {
      const arrondissement = parseInt(postcode.slice(-2));
      if (arrondissement == 1) {
        formattedCity = `Lyon 1er`;
      } else {
        formattedCity = `Lyon ${arrondissement}e`;
      }
    }
    return formattedCity;
  }

  /**
   * Search structure address based on data search WS
   * @param {searchQuery} data - Query address
   */
  public async searchAddress(data: {
    searchQuery: string;
  }): Promise<{ features: { geometry: { coordinates: number[] }; type: string; properties: unknown }[] }> {
    const reqBan = `https://download.data.grandlyon.com/geocoding/photon/api?q=${data.searchQuery}&lang=fr&limit=500&osm_tag=:!construction&osm_tag=:!bus_stop`;

    const reqBal = `https://download.data.grandlyon.com/geocoding/photon-bal/api?q=${data.searchQuery}&lang=fr&limit=500&osm_tag=:!construction&osm_tag=:!bus_stop`;

    const requestGroup = (
      url
    ): Promise<{ features: { geometry: { coordinates: number[] }; type: string; properties: unknown }[] }> =>
      new Promise((resolve) => {
        this.logger.debug(`Search request: ${encodeURI(url)}`, 'StructureService');
        return this.httpService
          .request({
            url: encodeURI(url),
            method: 'GET',
            headers: { 'Content-Type': 'application/json' },
          })
          .subscribe(
            (reply) => {
              this.logger.debug(`Search request response length : ${reply.data.features.length}`, 'StructureService');
              reply.data.features = reply.data.features
                .filter((doc) => doc.properties.postcode && doc.properties.postcode.match(depRegex))
                .sort((_a, b) => {
                  return b.properties.housenumber ? 1 : -1;
                });
              return resolve(reply.data);
            },
            (err) => {
              this.logger.error(`Search - Request error: ${err.config.url}`, 'StructureService');
              this.logger.error(err);
            }
          );
      });

    const [reqBanRes, reqBalRes] = await Promise.all([requestGroup(reqBan), requestGroup(reqBal)]);
    const mergedArray = [...reqBalRes['features'], ...reqBanRes['features']];
    const duplicateFreeArray = _.unionWith(mergedArray, function (a, b) {
      return (
        // excludes structures from photon-BAN if they share postcode && street or name && house number
        //  checking for name and street existence asserts that we will not compare 2 undefined fields
        a.properties.postcode === b.properties.postcode &&
        ((a.properties.name &&
          (a.properties.name === b.properties.name || a.properties.name === b.properties.street)) ||
          (a.properties.street &&
            (a.properties.street === b.properties.name || a.properties.street === b.properties.street))) &&
        (a.properties.housenumber === b.properties.housenumber ||
          (!a.properties.housenumber && !b.properties.housenumber))
      );
    });

    duplicateFreeArray.forEach((features) => {
      features.properties.city = this.getFormattedCity(features.properties.city, features.properties.postcode);
    });

    return {
      features: duplicateFreeArray,
    };
  }

  /**
   * Count every value occurrence of a given key
   * @param key structure key
   * @return [{id: 'key', count: 'value'}]
   */
  public async countByStructureKey(key: string, selected: { id: string; text: string }[]) {
    const uniqueElements = await this.structureModel.distinct(key).exec();
    return Promise.all(
      uniqueElements.map(async (value) => {
        const keyList: FilterQuery<DocumentDefinition<StructureDocument>>[] = [];
        keyList.push({
          [key]: { $elemMatch: { $eq: value } },
          deletedAt: { $exists: false },
          accountVerified: true,
        });
        if (selected && selected.length > 0) {
          for (const val of selected) {
            keyList.push({
              [val.text]: { $elemMatch: { $eq: val.id } },
              deletedAt: { $exists: false },
              accountVerified: true,
            });
          }
        }
        return {
          id: value,
          count: await this.structureModel
            .countDocuments({
              $and: keyList,
            })
            .exec(),
        };
      })
    );
  }

  public async getCoord(
    numero: string,
    address: string,
    commune: string,
    scope: string
  ): Promise<AxiosResponse<PhotonResponse>> {
    const req =
      `https://download.data.grandlyon.com/geocoding/${scope}/api?q=` +
      (numero == null ? '' : numero + ' ') +
      address +
      ' ' +
      commune;
    this.logger.debug('Print getCoord ' + req);

    return await lastValueFrom(this.httpService.get(encodeURI(req)));
  }

  /**
   * Set the structure to be deleted in 5 weeks
   * @param user IUser
   * @param structure StructureDocument
   */
  public async setToBeDeleted(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    structure.toBeDeletedAt = DateTime.local().plus({ weeks: 5 }).setZone('Europe/Paris').toString();
    structure.save();

    this.sendToBeDeletedNotification(user, structure);
    return structure;
  }

  /**
   * Send an email to structure owners (except the one who asked for the deletion) to inform the structure is will be deleted in 5 weeks
   * @param user User
   */
  public async sendToBeDeletedNotification(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureToBeDeleted.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureToBeDeleted.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      toBeDeletedAt: DateTime.fromISO(structure.toBeDeletedAt.toISOString())
        .plus({ days: 1 })
        .toLocaleString(DateTime.DATE_SHORT),
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  /**
   * Cancel the structure deletion when one of the structure owners had asked the structure for deletion
   * @param user IUser
   * @param structure StructureDocument
   */
  public async cancelDelete(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    if (!structure.deletedAt) {
      structure.toBeDeletedAt = null;
      structure.save();

      this.sendCancelDeleteNotification(user, structure);
    }
    return structure;
  }

  /**
   * Send an email to other structure owners to inform an owner cancelled the structure deletion
   * @param user User
   */
  public async sendCancelDeleteNotification(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureCancelDelete.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureCancelDelete.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  public async deleteOne(structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    this.structuresSearchService.deleteIndexStructure(structure);
    structure.structureType = null;
    if (structure.toBeDeletedAt) structure.toBeDeletedAt = null;
    structure.deletedAt = DateTime.local().setZone('Europe/Paris').toString();
    this.anonymizeStructure(structure).save();
    this.logger.debug(`delete structure : ${structure.structureName} (${structure._id})`);

    // Remove structure from owners (and check if there is a newly unattached user)
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      this.userService.removeFromStructureLinked(owner, structure._id);
    });
    this.sendAdminStructureNotification(
      structure,
      this.mailerService.config.templates.structureDeletionNotification.ejs,
      this.mailerService.config.templates.structureDeletionNotification.json
    );
    return structure;
  }

  public async sendAdminStructureNotification(
    structure: StructureDocument,
    templateLocation: string,
    jsonConfigLocation: string,
    user = null
  ) {
    const uniqueAdminEmails = [...new Set((await this.userService.getAdmins()).map((admin) => admin.email))].map(
      (item) => {
        return { email: item };
      }
    );

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);
    const html = await ejs.renderFile(ejsPath, {
      config,
      id: structure ? structure._id : 0,
      structureName: structure ? structure.structureName : '',
      structureAddress: structure
        ? `${structure.address.numero || ''} ${structure.address.street} ${structure.address.commune}`
        : '',
      structureDescription: structure ? structure.otherDescription : '',
      user: user,
    });
    this.mailerService.send(uniqueAdminEmails, jsonConfig.subject, html);
  }

  private anonymizeStructure(structure: StructureDocument): StructureDocument {
    structure.contactPhone = '';
    structure.contactMail = '';
    structure.facebook = null;
    structure.twitter = null;
    structure.instagram = null;
    structure.website = '';
    return structure;
  }

  @Cron(CronExpression.EVERY_DAY_AT_4AM)
  public async structuresTasksProcess(): Promise<void> {
    this.logger.log('structuresTasksProcess');

    await this.processToBeDeletedStructures().catch((error) => {
      this.logger.error(error);
    });

    await this.processOutdatedStructures().catch((error) => {
      this.logger.error(error);
    });
  }

  private async processToBeDeletedStructures(): Promise<void> {
    this.logger.debug('processToBeDeletedStructures');

    const structures = await this.structureModel
      .find()
      .where('toBeDeletedAt')
      .lte(DateTime.local())
      .where('deletedAt')
      .exists(false)
      .exec();

    const pendingStructures = (await this.userService.getPendingStructures()) as PendingStructureDto[];
    const structureIds = pendingStructures.map((pending) => pending.structureId);
    structures.forEach((structure) => {
      if (structureIds.includes(structure.id)) {
        this.logger.debug(`cancel structure soft-delete : ${structure.structureName} (${structure._id})`);
        this.structureModel.findByIdAndUpdate(new Types.ObjectId(structure.id), {
          toBeDeletedAt: null,
        });
      } else {
        this.deleteOne(structure);
      }
    });
  }

  private async processOutdatedStructures(): Promise<void> {
    this.logger.debug('processOutdatedStructures');
    const OUTDATED_MONTH_TO_CHECK = 6;
    const structureList = await this.findAll();
    // Get outdated structures
    const filteredList = structureList.filter((structure) => {
      const updatedDate = DateTime.fromISO(structure.updatedAt.toISOString());
      const diff = updatedDate.diffNow('month').months;
      if (Math.abs(diff) > OUTDATED_MONTH_TO_CHECK) {
        this.logger.debug(`found outdated structure : ${structure.structureName}`);
        return true;
      }
    });
    // Get owners of outdated structures
    const ownerList = await Promise.all(
      filteredList.map(async (structure) => {
        const owner = await this.userService.isStructureClaimed(structure._id.toString());
        if (owner) {
          return { structure: structure, owner: owner };
        }
      })
    );
    // Send email if possible and update user
    ownerList
      .filter((x) => x != undefined)
      .forEach(async (data) => {
        this.userService.findOne(data.owner.email).then((user) => {
          // If mail is already sent, do not resend
          if (user.structureOutdatedMailSent.includes(data.structure._id)) {
            return;
          } else {
            this.sendOutdatedEmailToUser(data.owner.email, data.structure.structureName, data.structure._id);
            user.structureOutdatedMailSent.push(data.structure._id);
            user.save();
          }
        });
      });
  }

  /**
   * Send an email to prevent outdated
   * a new account.
   * @param user User
   */
  private async sendOutdatedEmailToUser(userEmail: string, structureName: string, id: string): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureOutdatedInfo.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureOutdatedInfo.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      name: structureName,
      id: id,
    });
    this.mailerService.send(userEmail, jsonConfig.subject, html);
  }

  /**
   * Send an email to structure owners and admin in order to accept or decline a join request
   * @param user User
   */
  public async sendStructureJoinRequest(
    user: IUser,
    structure: StructureDocument,
    validationToken: string
  ): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureJoinRequest.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureJoinRequest.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      id: structure._id,
      userId: user._id,
      validationToken: validationToken,
    });
    const owners = await this.getOwners(structure._id);
    for (const owner of owners) {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    }
    const admins = await this.userService.getAdmins();
    for (const admin of admins) {
      this.mailerService.send(admin.email, jsonConfig.subject, html);
    }
  }

  private async getOwners(structureId: string): Promise<IUser[]> {
    // Get owners of outdated structures
    return this.userService.getStructureOwners(structureId);
  }

  public async updateAccountVerified(idStructure: string): Promise<Structure> {
    const structureLinked = await this.findOne(idStructure);
    const structure = new this.structureModel(structureLinked);
    if (!structure) {
      throw new HttpException('Invalid structure', HttpStatus.NOT_FOUND);
    }
    structure.accountVerified = true;
    structure.save();
    return structure;
  }

  public async findWithOwners(idStructure: string): Promise<{ structure: Structure; owners: IUser[] }> {
    const structure = await this.findOne(idStructure);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersDetails(structure._id);
    return { structure: structure, owners: owners };
  }

  public async reportStructureError(structureId: string, content: string) {
    const structure = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersDetails(structure._id);
    const admins = await this.userService.getAdmins();
    const emails = owners.map((owner) => owner.email).concat(admins.map((admin) => admin.email));
    const uniqueEmails = [...new Set(emails)];

    const emailsObject: { email: string }[] = uniqueEmails.map((item) => {
      return { email: item };
    });

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureErrorReport.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureErrorReport.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      content: content,
      id: structure._id,
      structureName: structure.structureName,
    });
    this.mailerService.send(emailsObject, jsonConfig.subject, html);
  }

  public async getAllUserCompletedStructures(users: IUser[]) {
    return Promise.all(
      users.map(async (user) => {
        return {
          id: user._id,
          createdAt: user.createdAt ? user.createdAt.toLocaleDateString('fr-FR') : '',
          surname: user.surname,
          name: user.name,
          email: user.email,
          phone: user.phone,
          job: user.job,
          employer: user.employer,
          structures: await Promise.all(
            user.structuresLink.map(async (id) => {
              return this.findOne(id.toHexString());
            })
          ),
        };
      })
    );
  }

  /**
   * Retrieve all structure pending for data.grandlyon consent sharing.
   * Case 1 : null value stands for 'no'
   * Case 2: a date stands for 'yes'
   * Case 3: no key in document when there is no reply to consent
   * @param users {IUser}
   * @returns {Structure[]}
   */
  public async getAllDataConsentPendingStructures(users: IUser): Promise<Structure[]> {
    const data = await Promise.all(
      users.structuresLink.map(async (structureId: Types.ObjectId) => {
        const structure = await this.findOne(structureId.toHexString());
        if (structure && structure.dataShareConsentDate === undefined) {
          return structure;
        }
      })
    );
    return data.filter((value) => value);
  }

  /**
   * Add the personal offer to the structure
   * @param structureId string
   * @param personalOfferDocument PersonalOfferDocument
   * @returns {Structure} structure with personal offer added
   */
  public async addPersonalOffer(
    structureId: string,
    personalOfferDocument: PersonalOfferDocument
  ): Promise<StructureDocument> {
    const structure: StructureDocument = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Structure not found for the personal offer attachment', HttpStatus.BAD_REQUEST);
    }
    if (
      structure.personalOffers &&
      structure.personalOffers.findIndex((personalOffer) => personalOffer._id === personalOfferDocument._id) > -1
    ) {
      throw new HttpException('Personal offer already exist in the structure', HttpStatus.BAD_REQUEST);
    } else {
      structure.personalOffers.push(personalOfferDocument);
    }
    await structure.save();
    return structure;
  }

  /**
   * Fetch espace coop and filter structures in 69XXX and structures with courriel and phone
   * @returns CNFSStructure[]
   */
  public getCNFSStructures(): Promise<Array<CNFSStructure>> {
    this.logger.debug(`get CNFS structures`);
    const obs = this.httpService.get('https://api.conseiller-numerique.gouv.fr/permanences').pipe(
      map((response: AxiosResponse<Array<CNFSStructure>>) =>
        response.data.filter((structure: CNFSStructure) => {
          if (structure.code_postal.substring(0, 2) === '69' && (structure.courriel || structure.telephone))
            return structure;
        })
      ),
      tap((response) => {
        this.logger.debug(`${response.length} CNFS rhone structures found`);
      })
    );

    return new Promise((resolve) => {
      obs.subscribe((data) => {
        resolve(data);
      });
    });
  }

  /**
   * Compare Resin & CNFS structures with email and phone, returns the structure when matching
   */
  public findMatchingStructure(resinStructure: StructureDocument, CNFSData: CNFSStructure[]): CNFSStructure | null {
    // This check is mandatory because if resin phone number is '' we should return FALSE instead of UNDEFINED
    const resinPhoneNumber = resinStructure.contactPhone?.split(' ').join('') || false;
    return CNFSData.find(
      (structure) =>
        structure.courriel === resinStructure.contactMail ||
        structure.telephone?.replace('+33', '0') === resinPhoneNumber
    );
  }

  /**
   * When a structure is created, fetch espace coop and tries to find a match to bind CNFSid to the resin structure.
   * @param resinStructureId string
   */
  public async bindOneCNFSStructure(resinStructureId: string): Promise<string> {
    this.logger.debug(`bind CNFS id to one resin structure`);
    const resinStructure = await this.findOne(resinStructureId);
    const CNFSData = await this.getCNFSStructures();
    if (!resinStructure) return 'Could not find structure';

    const matchingStructure = this.findMatchingStructure(resinStructure, CNFSData);
    if (matchingStructure && matchingStructure.prise_rdv) {
      try {
        const result = await this.setCNFSid(resinStructure.id, matchingStructure.structureId);
        if (result) {
          return 'A match has been found, updating CNFSid';
        }
      } catch (error) {
        this.logger.error(`Error updating CNFSid: ${error.message}`);
        return 'A match has been found, but there was an error updating CNFSid';
      }
    }
    this.logger.debug(`No match found`);
    return 'No match found';
  }

  /**
   * Fetch espace coop. If any structure matches, assign its CNFSid in our DB, if not assign ''
   * @returns "${number} of structures affected"
   */
  @Cron(CronExpression.EVERY_WEEK)
  public async bindCNFSids(): Promise<string> {
    this.logger.log('bind CNFS ids to resin structures process');
    const CNFSData = await this.getCNFSStructures();
    const resinStructures = await this.findAll();
    this.logger.debug(`${resinStructures.length} resin structures found`);

    let matchCount = 0;
    let mismatchCount = 0;

    for (const resinStructure of resinStructures) {
      const matchingStructure = this.findMatchingStructure(resinStructure, CNFSData);
      if (matchingStructure && matchingStructure.prise_rdv) {
        try {
          const result = await this.setCNFSid(resinStructure.id, matchingStructure.structureId);
          if (result) {
            matchCount++;
          }
        } catch (error) {
          this.logger.error(`Error while setting CNFS ID for structure ${resinStructure.id}: ${error.message}`);
        }
        // if a corresponding structure has been found but its email is different, we probably will need to update it here.
      } else {
        mismatchCount++;
        if (resinStructure.idCNFS) {
          // to reset the cnfs id field if it was previously set
          this.logger.warn(`Cleaning old CNFS id '${resinStructure.idCNFS}'`);
          const setCnfsToNull = Object.assign(resinStructure, { idCNFS: '' });
          this.structureModel.findByIdAndUpdate(new Types.ObjectId(resinStructure.id), setCnfsToNull).exec();
        }
      }
    }
    this.logger.log(
      `${matchCount} structures affected with an idCNFS (${mismatchCount} structures not matching espace coop ids)`
    );
    return `${matchCount} structures affected`;
  }

  /**
   * Update the idCNFS of a structure if it has meetings open on rdvs"
   */
  public async setCNFSid(idStructure: string, idCNFS: string): Promise<Structure> {
    const oldStructure = await this.findOne(idStructure);

    // no need for an update if the cnfs id match and is already known (in order to not mess with the structure activity recording when it was lastly updated for mail inquiries)
    if (oldStructure.idCNFS === idCNFS) {
      return null;
    }
    this.logger.log(`A resin structure '${idStructure}' matched with a CNFS one, updating its idCNFS '${idCNFS}'`);
    const deepClone = Object.assign(oldStructure, { idCNFS });
    return this.structureModel.findByIdAndUpdate(new Types.ObjectId(idStructure), deepClone).exec();
  }

  /**
   * Remove a personal offer from structure offers
   */
  public async removePersonalOffer(structure: StructureDocument, offerId: string): Promise<PersonalOfferDocument[]> {
    if (!structure) {
      throw new HttpException('Invalid structure', HttpStatus.NOT_FOUND);
    }
    structure.personalOffers = structure.personalOffers.filter(
      (personalOffer) => !personalOffer._id.equals(new Types.ObjectId(offerId))
    );
    await structure.save();
    return structure.personalOffers;
  }

  public async findByPersonalOfferId(id: string): Promise<StructureDocument> {
    this.logger.debug('findByPersonalOfferId');
    return this.structureModel
      .findOne({
        personalOffers: { $all: [new Types.ObjectId(id)] },
      })
      .populate('personalOffers')
      .exec();
  }

  /**
   * Set structure CTM Territory with data from Data Grand Lyon
   */
  public async setCTMs(): Promise<void> {
    this.logger.debug('setCTMs');
    const structures: StructureDocument[] = await this.findAll();
    for (const structure of structures) {
      this.setCtmTerritory(structure).then(async (updatedStructure) => {
        await this.structureModel
          .findByIdAndUpdate(new Types.ObjectId(structure._id), updatedStructure, { timestamps: false })
          .exec();
      });
    }
    return null;
  }
}
