import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import * as ejs from 'ejs';
import { Model, Types } from 'mongoose';
import { MailerService } from '../mailer/mailer.service';
import { CreateTempUserDto } from './dto/create-temp-user.dto';
import { ITempUser } from './temp-user.interface';
import { TempUser } from './temp-user.schema';

@Injectable()
export class TempUserService {
  private readonly logger = new Logger(TempUserService.name);

  constructor(
    private readonly mailerService: MailerService,
    @InjectModel(TempUser.name) private tempUserModel: Model<ITempUser>
  ) {}

  public async create(createTempUser: CreateTempUserDto, structureName: string): Promise<TempUser> {
    const userInDb = await this.findOne(createTempUser.email);
    if (userInDb) {
      throw new HttpException('User already exists', HttpStatus.BAD_REQUEST);
    }
    const createUser = await this.tempUserModel.create(createTempUser);

    this.logger.debug(`TempUsersService | tempUser created`);
    // Send email
    this.sendUserMail(createUser, structureName);
    return this.findOne(createTempUser.email);
  }

  public async findOne(mail: string): Promise<ITempUser> {
    return this.tempUserModel.findOne({ email: mail });
  }

  public async findById(id: string): Promise<TempUser> {
    return this.tempUserModel.findById(new Types.ObjectId(id));
  }

  public async delete(mail: string): Promise<TempUser> {
    const userInDb = await this.findOne(mail);
    if (!userInDb) {
      throw new HttpException('User does not exists', HttpStatus.BAD_REQUEST);
    }
    await this.tempUserModel.deleteOne({ email: mail });
    return userInDb;
  }

  public async updateStructureLinked(createTempUser: CreateTempUserDto): Promise<TempUser> {
    const userInDb = await this.tempUserModel
      .find({
        $and: [
          {
            email: createTempUser.email,
          },
          {
            structuresLink: { $in: [createTempUser.structuresLink[0]] },
          },
        ],
      })
      .exec();

    if (userInDb.length > 0) {
      throw new HttpException('User already linked', HttpStatus.UNPROCESSABLE_ENTITY);
    }
    return this.tempUserModel
      .findByIdAndUpdate(
        { email: createTempUser.email },
        { $push: { structuresLink: new Types.ObjectId(createTempUser.structuresLink[0]) } }
      )
      .exec();
  }

  /**
   * Send email in order to tell the user that an account is already fill with his structure info.
   * @param user User
   */
  public async sendUserMail(user: ITempUser, structureName: string, existingUser?: boolean): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(
      existingUser ? config.templates.userAddedToStructure.ejs : config.templates.tempUserRegistration.ejs
    );
    const jsonConfig = this.mailerService.loadJsonConfig(
      existingUser ? config.templates.userAddedToStructure.json : config.templates.tempUserRegistration.json
    );

    const html = await ejs.renderFile(ejsPath, {
      config,
      id: user._id,
      name: structureName,
    });
    this.mailerService.send(user.email, jsonConfig.subject, html);
  }

  public async getStructureTempUsers(structureId: string): Promise<ITempUser[]> {
    return this.tempUserModel
      .find({ structuresLink: new Types.ObjectId(structureId) })
      .select('email updatedAt')
      .exec();
  }

  public async removeFromStructureLinked(userEmail: string, idStructure: string): Promise<Types.ObjectId[]> {
    const user = await this.findOne(userEmail);
    this.logger.debug(`find user : ${JSON.stringify(user)}`);

    if (!user) {
      throw new HttpException('Invalid temp user', HttpStatus.NOT_FOUND);
    }
    if (!user.structuresLink.includes(new Types.ObjectId(idStructure))) {
      throw new HttpException("Temp user doesn't belong to this structure", HttpStatus.NOT_FOUND);
    }
    user.structuresLink = user.structuresLink.filter((structureId) => {
      return !structureId.equals(idStructure);
    });
    await user.save();
    return user.structuresLink;
  }
}
