import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import * as ejs from 'ejs';
import * as mongodb from 'mongodb';
import { Model, Types } from 'mongoose';
import { MergeEmployerDto } from '../../admin/dto/merge-employer.dto';
import { MailerService } from '../../mailer/mailer.service';
import { CreateEmployerDto } from '../dto/create-employer.dto';
import { IEmployer } from '../interfaces/employer.interface';
import { Employer, EmployerDocument } from '../schemas/employer.schema';
import { EmployerSearchService } from './employer-search.service';
import { UsersService } from './users.service';

@Injectable()
export class EmployerService {
  private readonly logger = new Logger(EmployerService.name);

  constructor(
    @InjectModel(Employer.name) private employerModel: Model<EmployerDocument>,
    private readonly userService: UsersService,
    private readonly mailerService: MailerService,
    private readonly employerSearchService: EmployerSearchService
  ) {}

  public async findAll(): Promise<Employer[]> {
    this.logger.debug('findAll');
    return this.employerModel.find().sort({ name: 1 });
  }

  public async findOne(idParam: string): Promise<EmployerDocument> {
    this.logger.debug('findOne');
    return this.employerModel.findById(new Types.ObjectId(idParam));
  }

  public async findAllUnvalidated(): Promise<Employer[]> {
    this.logger.debug('findAllUnvalidated');
    return this.employerModel.find({ validated: false });
  }

  public async findAllValidated(): Promise<IEmployer[]> {
    this.logger.debug(`findAllValidated`);
    return this.employerModel.find({ validated: true });
  }

  public async findByName(name: string): Promise<EmployerDocument> {
    this.logger.debug('findByName');
    return this.employerModel.findOne({ name });
  }

  public async create(employer: CreateEmployerDto, validated = false, sendAdminNotification = true): Promise<Employer> {
    this.logger.debug(`createEmployer: ${employer.name}`);
    await this.employerModel.create({ name: employer.name, validated });
    const document = await this.findByName(employer.name);
    this.employerSearchService.indexEmployer(document);
    if (sendAdminNotification) {
      this.sendAdminCreateNotification(
        document,
        this.mailerService.config.templates.adminEmployerCreate.ejs,
        this.mailerService.config.templates.adminEmployerCreate.json
      );
    }
    return document;
  }

  public async sendAdminCreateNotification(
    employer: EmployerDocument,
    templateLocation: string,
    jsonConfigLocation: string
  ): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);
    const html = await ejs.renderFile(ejsPath, {
      config,
      employerName: employer ? employer.name : '',
    });
    const admins = await this.userService.getAdmins();
    admins.forEach((admin) => {
      this.mailerService.send(admin.email, jsonConfig.subject, html);
    });
  }

  public async validate(employerId: string): Promise<Employer> {
    this.logger.debug(`validateEmployer: ${employerId}`);
    const employer = await this.employerModel.findById(new Types.ObjectId(employerId));
    if (employer) {
      employer.validated = true;
      employer.save();
      return employer;
    } else {
      throw new HttpException('Cannot validate employer. It might have been already validate', HttpStatus.NOT_FOUND);
    }
  }

  public async update(employerId: string, newEmployer: CreateEmployerDto): Promise<Employer> {
    this.logger.debug(`editEmployer: ${employerId}`);
    const employer = await this.employerModel.findById(new Types.ObjectId(employerId));
    if (employer) {
      employer.name = newEmployer.name;
      employer.save();
      return employer;
    } else {
      throw new HttpException('Cannot edit employer. It was not found in database.', HttpStatus.NOT_FOUND);
    }
  }

  public async mergeEmployer({
    sourceEmployerId,
    targetEmployerId,
  }: MergeEmployerDto): Promise<Employer | HttpException> {
    this.logger.debug(`mergeEmployer: ${sourceEmployerId} into ${targetEmployerId}`);
    const sourceEmployer = await this.employerModel.findById(new Types.ObjectId(sourceEmployerId));
    const targetEmployer = await this.employerModel.findById(new Types.ObjectId(targetEmployerId));
    if (targetEmployer && sourceEmployer) {
      this.userService.replaceEmployers(sourceEmployer, targetEmployer);
      if (!sourceEmployer.validated) {
        this.deleteInvalidEmployer(sourceEmployerId);
      }
      return targetEmployer;
    } else {
      throw new HttpException('Cannot operate on employer.', HttpStatus.NOT_FOUND);
    }
  }

  public async deleteInvalidEmployer(id: string): Promise<mongodb.DeleteResult> {
    this.logger.debug(`deleteInvalidEmployer: ${id}`);
    const document = await this.employerModel.findById(new Types.ObjectId(id));
    this.employerSearchService.deleteIndex(document, document._id);
    return this.employerModel.deleteOne({ _id: id });
  }

  public async deleteOneId(id: string): Promise<Employer> {
    const employer = await this.employerModel.findOne({ _id: id });
    if (!employer) {
      throw new HttpException('Invalid employer id', HttpStatus.BAD_REQUEST);
    }
    return employer.deleteOne();
  }

  public async deleteByName(name: string): Promise<mongodb.DeleteResult> {
    this.logger.debug(`deleteByname: ${name}`);
    const document = await this.findByName(name);
    this.employerSearchService.deleteIndex(document, document._id);
    return this.employerModel.deleteOne({ name });
  }

  // SEARCH
  public async searchByName(searchString: string): Promise<Employer[]> {
    this.logger.debug('searchByName');
    return this.employerSearchService.search(searchString);
  }

  public async initEmployerIndex(): Promise<Employer[]> {
    Logger.log('Reset structures indexes');
    await this.employerSearchService.dropIndex();
    await this.employerSearchService.createEmployerIndex();
    return this.populateES();
  }

  private async populateES(): Promise<Employer[]> {
    const employers = await this.findAll();
    await Promise.all(
      employers.map((employer: EmployerDocument) => {
        this.employerSearchService.indexEmployer(employer);
      })
    );
    return employers;
  }
}
