import { Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Model } from 'mongoose';
import { IUser } from '../interfaces/user.interface';
import { UserRegistrySearchBody } from '../interfaces/userRegistry-search-body.interface';
import { IUserRegistry, UserRegistryPaginatedResponse } from '../interfaces/userRegistry.interface';
import { Employer } from '../schemas/employer.schema';
import { Job } from '../schemas/job.schema';
import { User } from '../schemas/user.schema';
import { EmployerService } from './employer.service';
import { JobsService } from './jobs.service';
import { UserRegistrySearchService } from './userRegistry-search.service';

@Injectable()
export class UserRegistryService {
  private readonly logger = new Logger(UserRegistryService.name);
  constructor(
    @InjectModel(User.name) private userModel: Model<IUser>,
    private userRegistrySearchService: UserRegistrySearchService,
    private jobsService: JobsService,
    private employerService: EmployerService
  ) {}
  public maxPerPage = 20;

  public async findAllForIndexation(): Promise<IUserRegistry[]> {
    this.logger.debug('findAllForIndexation');
    return this.userModel
      .find({ email: { $ne: process.env.MAIL_CONTACT } })
      .where('emailVerified')
      .equals(true)
      .select('name surname _id job employer ')
      .populate('job employer')
      .collation({ locale: 'fr' })
      .sort({ surname: 1 })
      .exec();
  }

  public async countAllUserRegistry(): Promise<number> {
    this.logger.debug('countAllUserRegistry');
    return this.userModel
      .find({ email: { $ne: process.env.MAIL_CONTACT } })
      .where('emailVerified')
      .equals(true)
      .populate('employer')
      .populate('job')
      .select('name surname employer job _id ')
      .collation({ locale: 'fr' })
      .sort({ surname: 1 })
      .count()
      .exec();
  }

  public async findAllUserRegistry(page: number): Promise<UserRegistryPaginatedResponse> {
    this.logger.debug('findAllUserRegistry');
    const limit = this.maxPerPage * page;
    const count = await this.countAllUserRegistry();
    const docs = await this.userModel
      .find({ email: { $ne: process.env.MAIL_CONTACT } })
      .where('emailVerified')
      .equals(true)
      .populate('employer')
      .populate('job')
      .select('name surname employer job _id ')
      .limit(limit)
      .collation({ locale: 'fr' })
      .sort({ surname: 1 })
      .exec();
    return { count: count, docs: docs };
  }

  private callbackFilter(users: IUser[], employersList: Employer[], jobList: Job[]): IUser[] {
    this.logger.debug('callbackFilter');
    const jobNames: string[] = jobList.map((job) => job.name);
    const employersNames: string[] = employersList.map((e) => e.name);
    // For each filter list (job or employer), we'll filter the main user list in order to get only the user that have a job or employer contained in the filters array
    // For this, we use findIndex method on job/employer name

    if (employersList?.length && jobList?.length) {
      return users.filter(
        (user) =>
          jobNames.findIndex((n) => user.job?.name === n) > -1 &&
          employersNames.findIndex((n) => user.employer?.name === n) > -1
      );
    }
    if (employersList?.length) {
      return users.filter((user) => employersNames.findIndex((n) => user.employer?.name === n) > -1);
    }
    if (jobList?.length) {
      return users.filter((user) => jobNames.findIndex((n) => user.job?.name === n) > -1);
    }
    return users;
  }

  public async findUsersByNameEmployerOrJob(
    searchParam: string,
    page: number,
    jobs?: string[],
    employers?: string[]
  ): Promise<UserRegistryPaginatedResponse> {
    this.logger.debug('findUsersByNameEmployerOrJob');
    const results = await this.userRegistrySearchService.search(searchParam);
    const ids = results.map((result) => result.id);
    const limit = page * this.maxPerPage || this.maxPerPage;
    const jobList: Job[] = [];
    const employersList: Employer[] = [];
    if (jobs) {
      for (const job of jobs) {
        jobList.push(await this.jobsService.findByName(job));
      }
    }
    if (employers) {
      for (const employer of employers) {
        employersList.push(await this.employerService.findByName(employer));
      }
    }
    const resultsWithFilter = await this.userModel
      .find({
        _id: { $in: ids },
        email: { $ne: process.env.MAIL_CONTACT },
      })
      .where('emailVerified')
      .equals(true)
      .select('name surname employer job _id ')
      .populate('employer job')
      .collation({ locale: 'fr' })
      .sort({ surname: 1 })
      .exec()
      .then((res) => {
        return this.callbackFilter(res, employersList, jobList);
      });
    return { count: resultsWithFilter.length, docs: resultsWithFilter.splice(0, limit) };
  }

  // SEARCH
  public async searchByNameAndSurname(searchString: string): Promise<UserRegistrySearchBody[]> {
    this.logger.debug('searchByNameAndSurname');
    return this.userRegistrySearchService.search(searchString);
  }

  public async initUserRegistryIndex(): Promise<IUserRegistry[]> {
    Logger.log('Reset structures indexes');
    await this.userRegistrySearchService.dropIndex();
    await this.userRegistrySearchService.createUserRegistryIndex();
    return this.populateES();
  }

  private async populateES(): Promise<IUserRegistry[]> {
    const users = await this.findAllForIndexation();
    await Promise.all(
      users.map((user: IUserRegistry) => {
        this.userRegistrySearchService.indexUserRegistry(user);
      })
    );
    return users;
  }
}
